/* --COPYRIGHT--,BSD
 * Copyright (c) 2014, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * --/COPYRIGHT--*/
// *******************************************************************************
//	TIDM-ONEWIRE EEPROM emulation using one-wire communication protocol
//  R Brown
//  Texas Instruments Inc.
//  December 2015
//  Built with CCS Version: 6.0.1
//******************************************************************************

//*****************************************************************************
// Includes
//*****************************************************************************
#include <msp430.h> 

//*****************************************************************************
// Functions
//*****************************************************************************
void initializeMSP430(void);
unsigned char crc8(unsigned char[], unsigned char);
unsigned int crc16(unsigned char, unsigned int);
void sendPresence(void);
void processROM(unsigned int);
void processMEM(unsigned int);
void readROM(void);
void matchROM(void);
void searchROM(void);
void writeScratchpad(void);
void readScratchpad(void);
void copyScratchpad(void);
void readMemory(void);
unsigned char sendBit(unsigned char);
unsigned char receiveBit(void);

//*****************************************************************************
// Flags
//*****************************************************************************
unsigned char firstInterrupt = 1;
unsigned char matchSuccessful = 0;
unsigned char earlyReset = 0;
unsigned char rcFlag = 0;
unsigned char timerOverflow = 0;

//*****************************************************************************
// Variables
//*****************************************************************************
unsigned int timerValue = 0;
unsigned int writeValue = 0;
unsigned int loop = 0;
unsigned int intBitMask = 0;
unsigned char bitSend = 0;
unsigned char bitCount = 0;
unsigned char byteCounter = 0;
unsigned char countInterrupts = 0;
unsigned char charBitMask = 0;
// Rom & scratchpad
unsigned char registrationNumber[8] = { 0 };
unsigned char scratchpad[32] = { 0 };
// Address registers
unsigned int targetAddress = 0;
unsigned int endAddressDataStatus = 0;

//*****************************************************************************
// Constants
//*****************************************************************************
const char family = 0x23;

//*****************************************************************************
// Defines
//*****************************************************************************
//Pointers to device-specific ID
#define LOTID  *((unsigned long *)(0x1A0A))   	// Lot/Wafer ID device descriptor
#define DIEX  *((unsigned int *)(0x1A0E))   	// Die x device descriptor
#define DIEY  *((unsigned int *)(0x1A10))   	// Die y device descriptor
#define MEMORY_SIZE 512							// Only allowable change to code, max value of 2^15 (32768)
//#define USE_PARASITIC_POWER					// If defined, only a maximum of 64 bytes of memory data will be sent at a time
//#define PARASITE_BYTES 64						// If USE_PARASITIC_POWER is defined, specify the number of bytes to send

//*****************************************************************************
// FRAM memory allocation
//*****************************************************************************
#pragma PERSISTENT (memory)
unsigned char memory[MEMORY_SIZE] = { 0 };

//*****************************************************************************
// Main function
// Initializes MSP430 peripherals, creates the 64-bit registration number,
// and reads input pulses to decipher between Memory and ROM function commands
//*****************************************************************************
int main(void) {
	initializeMSP430();										// Initialization of registers & peripherals

	// Initialize the registration number
	registrationNumber[0] = family;							// Family number
	registrationNumber[1] = LOTID;
	registrationNumber[2] = LOTID >> 8; 					// Least two bytes of Lot ID
	registrationNumber[3] = LOTID >> 16;
	registrationNumber[4] = LOTID >> 24;					// Greater two bytes of Lot ID
	registrationNumber[5] = DIEX;
	registrationNumber[6] = DIEY;							// First bytes of Die X & Y
	registrationNumber[7] = crc8(registrationNumber, 7);	// Calculated CRC

	while (1) {
		if (!earlyReset)									// If early reset flag high, do not enter LPM
		{
			TA1CTL |= MC__UP | TACLR;          				// Timer A1 Up mode, reset counter
			TA1CCR0 = 16000;								// Load timer for pause value
			timerOverflow = 0;								// Reset timer overflow
			__bis_SR_register(LPM0_bits | GIE);
			TA1CTL &= ~MC__UP;								// Stop Timer A1
			if(timerOverflow)								// If timer overflow occurs, the next pulse should be a reset
			{
				TA0CTL &= ~MC__CONTINUOUS;					// Stop TA0, which uses SMCLK
				P1SEL0 = 0x00;								// Change P1.7 selection to GPIO
				P1IFG = 0x00;								// Reset P1 interrupt flags
				P1IE = BIT7;								// P1.7 interrupt enable
				__bis_SR_register(LPM4_bits | GIE);			// Safe to enter a more efficient LPM to save power
				P1SEL0 = BIT7;								// TA0.CCI1A peripheral selection
				P1IE = 0x00;								// Disable P1.7 interrupt
				TA0CTL |= MC__CONTINUOUS | TACLR;			// Reset timer counter
				__bis_SR_register(LPM0_bits | GIE);
			}
		}
		if ((timerValue >= 7200 && timerValue <= 9600) || earlyReset)// Reset pulse detected between 450 and 600 us @ 16 MHz
		{													// or early reset from elsewhere in the code logic
			earlyReset = 0;									// Reset early reset flag
			sendPresence();									// Respond with presence pulse
			bitCount = 0;									// Reset bit count
			writeValue = 0;									// Reset write value
		}
		else if (timerValue >= 800 && timerValue <= 1920)	// Write logic 0 detected between 50 and 120 us @ 16 MHz
			bitCount++;										// No need to set a bit, move to next MSB
		else if (timerValue <= 240)							// Write logic 1 detected between 0 and 15 us @ 16 MHz
		{
			writeValue |= 1 << bitCount;					// Set 1 to proper writeValue bit
			bitCount++;										// Move to next MSB
		}
		if (bitCount == 8) 									// Has a full byte been sent?
		{
			bitCount = 0;									// Reset bit counter
			if (matchSuccessful == 0)
				processROM(writeValue);						// If no match condition then process the ROM command
			else if (matchSuccessful == 1)
				processMEM(writeValue);						// If match condition successful then process the memory function command
			writeValue = 0;									// Reset writeValue variable
			if (countInterrupts % 2)						// If number of interrupts is odd, need to reset
			{
				countInterrupts = 0;						// Reset interrupt counter
				firstInterrupt = 1;							// Reset interrupt flag
			}
		}
	}
}

//*****************************************************************************
// Initialize MSP430 function
// Stops the WDT, sets all ports for the lowest power settings possible,
// increases the DCO to 16 MHz, and initializes TA1 and TA0/CCI1A for pulse logic
//*****************************************************************************

void initializeMSP430() {
	WDTCTL = WDTPW | WDTHOLD;		// Stop WDT

	// Configure GPIO
	P1DIR = 0x7F;					// Set P1.7 to input direction
	P1OUT = 0x00;					// Set P1.7 pullup resistor off
	P1SEL0 = BIT7;					// TA0.CCI1A peripheral selection
	P1IES = BIT7;					// Set P1.7 interrupt from high to low transition

	P2DIR = 0xFF;					// Initialize all ports to output GPIO mode
	P3DIR = 0xFF;					// for lowest power consumption possible
	P4DIR = 0xFF;
	P5DIR = 0xFF;
	P6DIR = 0xFF;
	P7DIR = 0xFF;
	P8DIR = 0xFF;

	P2OUT = 0x00;
	P3OUT = 0x00;
	P4OUT = 0x00;
	P5OUT = 0x00;
	P6OUT = 0x00;
	P7OUT = 0x00;
	P8OUT = 0x00;

	P2SEL0 = 0x00;
	P3SEL0 = 0x00;
	P4SEL0 = 0x00;
	P5SEL0 = 0x00;
	P6SEL0 = 0x00;
	P7SEL0 = 0x00;
	P8SEL0 = 0x00;

	// Disable the GPIO power-on default high-impedance mode to activate
	// previously configured port settings
	PM5CTL0 &= ~LOCKLPM5;

	// Configure one FRAM waitstate as required by the device datasheet for MCLK
	// operation beyond 8MHz _before_ configuring the clock system.
	FRCTL0 = FRCTLPW | NWAITS_1;

	// Clock System Setup
    __bis_SR_register(SCG0);                           // disable FLL
    CSCTL3 |= SELREF__REFOCLK;                         // Set REFO as FLL reference source
    CSCTL0 = 0;                                        // clear DCO and MOD registers
    CSCTL1 &= ~(DCORSEL_7);                            // Clear DCO frequency select bits first
    CSCTL1 |= DCORSEL_5;                               // Set DCO = 16MHz
    CSCTL2 = FLLD_0 + 487;                             // DCOCLKDIV = 16MHz
    __delay_cycles(3);
    __bic_SR_register(SCG0);                           // enable FLL
    while(CSCTL7 & (FLLUNLOCK0 | FLLUNLOCK1));         // FLL locked

    CSCTL4 |= SELMS__DCOCLKDIV | SELA__REFOCLK;        // set default REFO(~32768Hz) as ACLK source, ACLK = 32768Hz
                                                       // default DCOCLKDIV as MCLK and SMCLK source

	// TA1 will handle the early reset logic
	TA1CCTL0 = CCIE;                          			// TACCR1 interrupt enabled
	TA1CTL = TASSEL__SMCLK | MC__STOP;          		// SMCLK, Stopped for now

	// TA0 counts the pulse length on the communication line
	TA0CCTL1 = CM_3 | CCIS_0 | SCS | CAP | CCIE;
	// Capture rising/falling edges,
	// Use CCI1A=P1.7/TA0.1,
	// Synchronous capture,
	// Enable capture mode,
	// Enable capture interrupt
	TA0CCTL1 &= ~COV;							// Reset capture overflow
	TA0CTL = TASSEL__SMCLK | MC__CONTINUOUS;  	// SMCLK, Continuous mode
	TA0CTL |= TACLR;							// Reset counter
}

//*****************************************************************************
// Send presence function
// If a reset pulse is detected, the MSP430 responds with a presence pulse
//*****************************************************************************
void sendPresence() {
	TA0CTL &= ~MC__CONTINUOUS;					// Stop timer
	P1OUT |= BIT7;								// Set P1.7 output latch
	P1SEL0 &= ~BIT7;							// Set P1.7 to I/O mode
	P1DIR |= BIT7;								// Set P1.7 to output direction
	TA1CTL |= MC__UP | TACLR;          			// Timer A1 Up mode, reset counter
	TA1CCR0 = 480;								// Delay for 30 us @ 16 MHz
	__bis_SR_register(LPM0_bits | GIE);			// Wait for Timer A1 interrupt
	P1OUT &= ~BIT7;								// Drive one-wire line low
	TA1CTL |= TACLR;
	TA1CCR0 = 1600;								// Delay 100 us @ 16 MHz
	__bis_SR_register(LPM0_bits | GIE);			// Wait for Timer A1 interrupt
	TA1CTL &= ~MC__UP;							// Stop Timer A1
	P1OUT |= BIT7;								// Release one-wire line
	P1DIR &= ~BIT7;							  	// Set P1.7 to input direction
	P1SEL0 |= BIT7;							  	// TA0.CCI1A peripheral selection
	TA0CTL |= MC__CONTINUOUS | TACLR;			// Clear counter & restart timer
	TA0CCTL1 &= ~CCIFG;
}

//*****************************************************************************
// Process ROM function
// Contains all of the necessary ROM function commands as well as a few flag updates
// \romCommand	Value of the ROM function command to execute
//*****************************************************************************
void processROM(unsigned int romCommand) {
	switch (romCommand) {
	case 0x33:								// Read ROM
		readROM();
		break;
	case 0x55:								// Match ROM
		matchROM();
		break;
	case 0xF0:								// Search ROM
		searchROM();
		break;
	case 0xCC:								// Skip ROM, mark as successful
		matchSuccessful = 1;
		break;
	case 0xA5:								// Resume, mark status according to previous match command
		if (rcFlag) matchSuccessful = 1;
		break;
	default:
		break;
	}
}

//*****************************************************************************
// Process memory function
// Contains all of the necessary memory function commands
// \memoryCommand	Value of the memory function command to execute
//*****************************************************************************
void processMEM(unsigned int memoryCommand) {
	matchSuccessful = 0;					// No longer a match
	switch (memoryCommand) {
	case 0x0F:								// Write Scratchpad
		writeScratchpad();
		break;
	case 0xAA:								// Read Scratchpad
		readScratchpad();
		break;
	case 0x55:								// Copy Scratchpad
		copyScratchpad();
		break;
	case 0xF0:								// Read Memory
		readMemory();
		break;
	default:
		break;
	}
}

//*****************************************************************************
// Read ROM function
// Sends the 8-byte ROM value to the master device
//*****************************************************************************
void readROM() {
	for (loop = 0; loop < 8; loop++){		// Loop to read the 8 byte ROM value if only one slave is present
		for (charBitMask = 0x01; charBitMask; charBitMask <<= 1) {
			bitSend = (charBitMask & registrationNumber[loop]) ? 1 : 0;
			if (sendBit(bitSend) == 1) {
				earlyReset = 1;
				return;
			}
		}
	}
}

//*****************************************************************************
// Match ROM function
// If the ROM value sent by the master matches the MSP430 device's ROM number,
// get ready to receive memory commands.  Otherwise ignore the incoming commands
//*****************************************************************************
void matchROM() {
	byteCounter = 0;
	bitCount = 0;
	unsigned char readROM[8] = { 0 };
	while (byteCounter < 8) {
		__bis_SR_register(LPM0_bits | GIE);
		if (timerValue >= 7200 && timerValue <= 9600) 		// Reset pulse detected between 450 and 600 us @ 16 MHz
		{
			earlyReset = 1;									// Early reset pulse, restart
			return;
		}
		else if (timerValue >= 800 && timerValue <= 1920)	// Write logic 0 detected between 50 and 120 us @ 16 MHz
			bitCount++;										// No need to set a bit, move to next MSB
		else if (timerValue <= 240)							// Write logic 1 detected between 0 and 15 us @ 16 MHz
		{
			readROM[byteCounter] |= 1 << bitCount;			// Set 1 to proper writeValue bit
			bitCount++;										// Move to next MSB
		}
		if (bitCount == 8) 									// Has a full byte been sent?
		{
			bitCount = 0;									// Reset bit counter
			if (readROM[byteCounter] != registrationNumber[byteCounter]) {
				matchSuccessful = 2;							// If not a match, update failed status and return
				rcFlag = 0;
				return;
			}
			byteCounter++;
		}
	}
	matchSuccessful = 1;									// Otherwise all bytes match, mark successful
	rcFlag = 1;
}

//*****************************************************************************
// Search ROM function
// MSP430 device sends the ROM value so long as the master keeps requesting it.
// Otherwise, it halts to allow other one-wire devices to send their ROM value.
//*****************************************************************************
void searchROM() {
	for (loop = 0; loop < 8; loop++)						// Loop to read 8 byte ROM value of multiple 1-wire slaves
	{
		for (charBitMask = 0x01; charBitMask; charBitMask <<= 1) {
			bitSend = (charBitMask & registrationNumber[loop]) ? 1 : 0;
			if (sendBit(bitSend) == 1) {
				earlyReset = 1;
				return;
			}
			if (sendBit(!bitSend) == 1) {
				earlyReset = 1;
				return;
			}
			if (receiveBit() != bitSend)
				return;
		}
	}
}

//*****************************************************************************
// Write Scratchpad function
// Receives data bytes from the master and stores them into the scratchpad.
// Computes a CRC16 value if the scratchpad has been filled
//*****************************************************************************
void writeScratchpad() {
	unsigned char bytesSent[34] = { 0 };
	unsigned int crc16Value = 0;
	bitCount = 0;
	byteCounter = 0;
	endAddressDataStatus = 1 << 6;								// Set PF bit to start off
	targetAddress = 0;
	crc16Value = crc16(0x0F, crc16Value);						// 0x0F is command code for writeScratchpad

	while (1) {
		__bis_SR_register(LPM0_bits | GIE);
		if (timerValue >= 7200 && timerValue <= 9600) 			// Reset pulse detected between 450 and 600 us @ 16 MHz
		{
			earlyReset = 1;										// Early reset pulse, restart
			if (bitCount == 0 && byteCounter > 1) 				// New byte and past the target address? Update end address
				endAddressDataStatus = (targetAddress & 0x01F)
						+ (byteCounter - 2);
			else if (bitCount != 0 && byteCounter > 1)			// Past target address but interrupted byte? Update end address with partial byte flag set
				endAddressDataStatus = (targetAddress & 0x01F)
						+ (byteCounter - 2) + 0x020;
			return;
		}
		else if (timerValue >= 800 && timerValue <= 1920)		// Write logic 0 detected between 50 and 120 us @ 16 MHz
			bitCount++;											// No need to set a bit, move to next MSB
		else if (timerValue <= 240)								// Write logic 1 detected between 0 and 15 us @ 16 MHz
		{
			if (byteCounter <= 1)
				targetAddress |= 1 << (bitCount + byteCounter * 8);	// If first two bytes set 1 to proper targetAddress bit
			else
				scratchpad[byteCounter - 2] |= 1 << bitCount;	// Else set 1 to proper scratchpad bit
			bytesSent[byteCounter] |= 1 << bitCount;			// This is for the CRC16 generation
			bitCount++;											// Move to next MSB
		}
		if (bitCount == 8) 										// Has a full byte been sent?
		{
			bitCount = 0;										// Reset bit counter
			byteCounter++;
			crc16Value = crc16(bytesSent[byteCounter - 1], crc16Value);	// Calculate next crc16 iteration
		}
		if ((targetAddress & 0x01F) + (byteCounter - 2) == 32)	// Has the scratchpad filled up?
		{
			endAddressDataStatus = 0x01F;						// Set end address value
			for (intBitMask = 0x01; intBitMask; intBitMask <<= 1) {
				bitSend = (intBitMask & crc16Value) ? 1 : 0;
				if (sendBit(!bitSend) == 1) {
					earlyReset = 1;
					return;
				}
			}
			return;
		}

	}
}

//*****************************************************************************
// Read scratchpad function
// Sends the scratchpad contents to the master device for verification
//*****************************************************************************
void readScratchpad() {

	for (intBitMask = 0x01; intBitMask; intBitMask <<= 1)				// Send the target address
	{
		if (sendBit((intBitMask & targetAddress) ? 1 : 0) == 1) {
			earlyReset = 1;
			return;
		}
	}
	for (charBitMask = 0x01; charBitMask; charBitMask <<= 1)			// Send the end address and data status flags
	{
		if (sendBit((charBitMask & endAddressDataStatus) ? 1 : 0) == 1) {
			earlyReset = 1;
			return;
		}
	}
	for (loop = 0; loop < (0x020 - (targetAddress & 0x01F)); loop++)	// Send the scratchpad contents
	{
		for (charBitMask = 0x01; charBitMask; charBitMask <<= 1) {
			bitSend = (charBitMask & scratchpad[loop]) ? 1 : 0;
			if (sendBit(bitSend) == 1) {
				earlyReset = 1;
				return;
			}
		}
	}
}

//*****************************************************************************
// Copy scratchpad function
// Saves the scratchpad contents into allocated FRAM memory if the master sends
// the correct start and end addresses.  Then returns alternating 0 and 1 bits
// to indicate that the save has been completed
//*****************************************************************************
void copyScratchpad() {
	byteCounter = 0;
	bitCount = 0;
	unsigned int checkTargetAddress = 0;
	unsigned int checkEndAddress = 0;

	while (1) {
		__bis_SR_register(LPM0_bits | GIE);
		if (timerValue >= 7200 && timerValue <= 9600) 		// Reset pulse detected between 450 and 600 us @ 16 MHz
		{
			earlyReset = 1;									// Early reset pulse, restart
			return;
		}
		else if (timerValue >= 800 && timerValue <= 1920)	// Write logic 0 detected between 50 and 120 us @ 16 MHz
			bitCount++;										// No need to set a bit, move to next MSB
		else if (timerValue <= 240)							// Write logic 1 detected between 0 and 15 us @ 16 MHz
		{
			if (byteCounter <= 1)
				checkTargetAddress |= 1 << (bitCount + byteCounter * 8);	// If first two bytes set 1 to proper targetAddress bit
			else
				checkEndAddress |= 1 << bitCount;			// Else set 1 to proper scratchpad bit
			bitCount++;										// Move to next MSB
		}
		if (bitCount == 8) 									// Has a full byte been sent?
		{
			bitCount = 0;									// Reset bit counter
			byteCounter++;
		}
		if (byteCounter == 3) {
			if (checkTargetAddress != targetAddress || checkEndAddress != endAddressDataStatus)
				return;
			else											// Only copy from scratchpad to memory if addresses align correctly
			{
				SYSCFG0 &= ~PFWP;               			// Unlock FRAM for write
				for (loop = 0; loop <= ((endAddressDataStatus & 0x01F) - (targetAddress & 0x01F)); loop++)
					memory[targetAddress + loop] = scratchpad[loop];
				SYSCFG0 |= PFWP;               				// Lock FRAM
				endAddressDataStatus |= 0x80;				// Update authorization accepted bit
				for (;;) {
					if (sendBit(0) == 1) {
						earlyReset = 1;
						return;
					}										//Send 0xAA repeatedly until reset condition present
					if (sendBit(1) == 1) {
						earlyReset = 1;
						return;
					}
				}
			}
		}
	}
}

//*****************************************************************************
// Read memory function
// Sends the memory contents to the master, the length of which depends on the
// settings determined by the type of power sourcing the MSP430
//*****************************************************************************
void readMemory() {
	unsigned int startAddress = 0;
	byteCounter = 0;
	bitCount = 0;

	while (1) {
		__bis_SR_register(LPM0_bits | GIE);
		if (timerValue >= 7200 && timerValue <= 9600) 			// Reset pulse detected between 450 and 600 us @ 16 MHz
		{
			earlyReset = 1;										// Early reset pulse, restart
			return;
		}
		else if (timerValue >= 800 && timerValue <= 1920)		// Write logic 0 detected between 50 and 120 us @ 16 MHz
			bitCount++;											// No need to set a bit, move to next MSB
		else if (timerValue <= 240)								// Write logic 1 detected between 0 and 15 us @ 16 MHz
		{
			startAddress |= 1 << (bitCount + byteCounter * 8);	// If first two bytes set 1 to proper startAddress bit
			bitCount++;											// Move to next MSB
		}
		if (bitCount == 8) 										// Has a full byte been sent?
		{
			bitCount = 0;										// Reset bit counter
			byteCounter++;
		}
		if (byteCounter == 2)									// Past the startAddress value, ready to send data
		{
#ifndef USE_PARASITIC_POWER										// If parasitic power is not being used, send full contents
			for (loop = 0; loop < MEMORY_SIZE - startAddress; loop++)	// Loop to send memory contents from start to end of memory
			{
				for (charBitMask = 0x01; charBitMask; charBitMask <<= 1) {
					bitSend =
							(charBitMask & memory[startAddress + loop]) ? 1 : 0;
					if (sendBit(bitSend) == 1) {
						earlyReset = 1;
						return;
					}
				}
			}
#else															// If parasitic power is being used, send a set number of bytes
			for (loop = 0; loop < PARASITE_BYTES; loop++)		// Send a chunk of memory contents within the parasite power's capabilities
			{
				for(charBitMask = 0x01; charBitMask; charBitMask <<= 1)
				{
					bitSend = (charBitMask & memory[startAddress + loop])?1:0;
					if(sendBit(bitSend) == 1) {earlyReset = 1;return;}
					else if ((startAddress + loop) >= (MEMORY_SIZE - 1)) return;// Reached end of memory before pre-determined number
				}																// of bytes was reached, therefore return early
			}
#endif
		}
	}
}

//*****************************************************************************
// Send bit function
// Sends a bit value to the master device
// \bit		Bit value to be send to the master
// \return	Bit was either sucessfully sent or a reset occurred
//*****************************************************************************
unsigned char sendBit(unsigned char bit) {
	TA1CTL |= MC__UP | TACLR;          			// Timer A1 Up mode, reset counter
	TA1CCR0 = 3000;								// Load timer for reset value
	firstInterrupt = 0;							// Force exit from LPM on next ISR
	timerOverflow = 0;							// Reset timer overflow
	__bis_SR_register(LPM0_bits | GIE);

	if (timerOverflow)							// Line has been low for too long, must be a reset condition
	{
		TA1CTL &= ~MC__UP;						// Stop Timer A1
		__bis_SR_register(LPM0_bits | GIE);		// Finish out the reset
		countInterrupts = 0;					// Reset interrupt counter
		firstInterrupt = 1;						// Reset interrupt flag

		TA0CTL |= TACLR;						// Clear counter
		return 1;
	}
	else if (bit == 0)							// If a zero value, drive pin low
	{
		P1DIR |= BIT7;							// Set P1.7 to output direction, drive low
		P1OUT &= ~BIT7;							// Set P1.7 output latch
		P1SEL0 &= ~BIT7;						// Set P1.7 to I/O mode
	}
	TA1CTL |= TACLR;
	TA1CCR0 = 480;								// Delay for 30 us @ 16 MHz
	__bis_SR_register(LPM0_bits | GIE);			// Wait for Timer A1 interrupt
	TA1CTL &= ~MC__UP;							// Stop Timer A1

	P1SEL0 |= BIT7;							  	// TA0.CCI1A peripheral selection
	P1DIR &= ~BIT7;							  	// Set P1.7 to input direction
	P1OUT |= BIT7;								// Release one-wire line

	countInterrupts = 0;						// Reset interrupt counter
	firstInterrupt = 1;							// Reset interrupt flag

	TA0CTL |= TACLR;							// Clear counter
	return 0;
}

//*****************************************************************************
// Receive bit function
// Receives a bit from the master device
// \return	Value of the receive bit
//*****************************************************************************
unsigned char receiveBit() {
	unsigned char bit = 0;
	__bis_SR_register(LPM0_bits | GIE);
	if (timerValue >= 7200 && timerValue <= 9600)	// Reset pulse detected between 450 and 600 us @ 16 MHz
	{
		earlyReset = 1;								// Early reset condition set
		return 2;									// Return a non-bit value
	}
	else if (timerValue <= 240) bit = 1;			// Write logic 1 detected between 0 and 15 us @ 16 MHz
	return bit;
}

//*****************************************************************************
// CRC8 function
// Calculates the CRC8 value for the MSB of the registration number
// \addr[]	The least significant 7 bytes of the registration number
// \length	Number of bytes to be checked by the CRC
// \return	CRC8 byte added to the end of the registration number
//*****************************************************************************
unsigned char crc8(unsigned char addr[], unsigned char length) {
	unsigned char crc8 = 0;

	while (length--) {
		unsigned char inbyte = *addr++;
		for (loop = 8; loop; loop--) {
			unsigned char mix = (crc8 ^ inbyte) & 0x01;
			crc8 >>= 1;
			if (mix)
				crc8 ^= 0x8C;
			inbyte >>= 1;
		}
	}
	return crc8;
}

//*****************************************************************************
// CRC16 function
// Calculates the next iteration of the CRC16 value based on the previous input
// \input	New byte to be appended to the CRC16
// \crc16	Previous CRC16 value
// \return	Updated CRC16 value
//*****************************************************************************
unsigned int crc16(unsigned char input, unsigned int crc16) {
	static const unsigned char oddparity[16] = { 0, 1, 1, 0, 1, 0, 0, 1, 1, 0,
			0, 1, 0, 1, 1, 0 };

	unsigned int cdata = input;
	cdata = (cdata ^ crc16) & 0xff;
	crc16 >>= 8;

	if (oddparity[cdata & 0x0F] ^ oddparity[cdata >> 4])
		crc16 ^= 0xC001;

	cdata <<= 6;
	crc16 ^= cdata;
	cdata <<= 1;
	crc16 ^= cdata;

	return crc16;
}

//*****************************************************************************
// Timer0_A3 CC1-2, TA Interrupt Handler
// Measures the length of time that the master holds the one-wire line low
//*****************************************************************************
#if defined(__TI_COMPILER_VERSION__) || defined(__IAR_SYSTEMS_ICC__)
#pragma vector = TIMER0_A1_VECTOR
__interrupt void Timer0_A1_ISR(void)
#elif defined(__GNUC__)
void __attribute__ ((interrupt(TIMER0_A1_VECTOR))) Timer0_A1_ISR (void)
#else
#error Compiler not supported!
#endif
{
	switch (__even_in_range(TA0IV, TA0IV_TAIFG)) {
	case TA0IV_TACCR1:
		if (!firstInterrupt) 						// If no first interrupt flag, action required
		{
			timerValue = TA0CCR1;					// Record pulse length
			__bic_SR_register_on_exit(LPM4_bits);	// Exit LPM(0-4) to service pulse
			firstInterrupt = 1;
		}
		else firstInterrupt = 0;					// Reset flag after first interrupt
		TA0CTL |= TACLR;				  			// Reset timer counter
		countInterrupts++;
		break;
	case TA0IV_TACCR2:
		break;
	case TA0IV_TAIFG:
		break;
	default:
		break;
	}
}

//*****************************************************************************
// Timer1_A0 interrupt service routine
// Forces an exit from LPM0 if the master is initiating an early reset
// or indicates that standby mode should be initiated
//*****************************************************************************
#if defined(__TI_COMPILER_VERSION__) || defined(__IAR_SYSTEMS_ICC__)
#pragma vector = TIMER1_A0_VECTOR
__interrupt void Timer1_A0_ISR(void)
#elif defined(__GNUC__)
void __attribute__ ((interrupt(TIMER1_A0_VECTOR))) Timer1_A0_ISR (void)
#else
#error Compiler not supported!
#endif
{
	timerOverflow = 1;								// Set timerOverflow flag
	__bic_SR_register_on_exit(LPM4_bits);			// Exit LPM(0-4) to service overflow
}

//*****************************************************************************
// Port 1 interrupt service routine
// Forces an exit from LPM if the master begins a reset pulse
//*****************************************************************************
#if defined(__TI_COMPILER_VERSION__) || defined(__IAR_SYSTEMS_ICC__)
#pragma vector=PORT1_VECTOR
__interrupt void Port_1(void)
#elif defined(__GNUC__)
void __attribute__ ((interrupt(PORT1_VECTOR))) Port_1 (void)
#else
#error Compiler not supported!
#endif
{
	firstInterrupt = 1;								// Reset flag after first interrupt
	P1IFG &= ~BIT7;                         		// Clear P1.7 IFG
	__bic_SR_register_on_exit(LPM4_bits);     		// Exit LPM(0-4)
}
